@@ -50,9 +50,9 @@
 #include <dirent.h>
 #include <syslog.h>
 #include <grp.h>
+#include <systemd/sd-daemon.h>
 
 #include "rtkit.h"
-#include "sd-daemon.h"
 
 #ifndef __linux__
 #error "This stuff only works on Linux!"
@@ -67,48 +67,8 @@
 #define RLIMIT_RTTIME 15
 #endif
 
-#define INTROSPECT_XML                                                  \
-        DBUS_INTROSPECT_1_0_XML_DOCTYPE_DECL_NODE                       \
-        "<node>\n"                                                      \
-        "        <interface name=\"org.freedesktop.RealtimeKit1\">\n"   \
-        "                <method name=\"MakeThreadRealtime\">\n"        \
-        "                        <arg name=\"thread\" type=\"t\" direction=\"in\"/>\n" \
-        "                        <arg name=\"priority\" type=\"u\" direction=\"in\"/>\n" \
-        "                </method>\n"                                   \
-        "                <method name=\"MakeThreadRealtimeWithPID\">\n"        \
-        "                        <arg name=\"process\" type=\"t\" direction=\"in\"/>\n" \
-        "                        <arg name=\"thread\" type=\"t\" direction=\"in\"/>\n" \
-        "                        <arg name=\"priority\" type=\"u\" direction=\"in\"/>\n" \
-        "                </method>\n"                                   \
-        "                <method name=\"MakeThreadHighPriority\">\n"          \
-        "                        <arg name=\"thread\" type=\"t\" direction=\"in\"/>\n" \
-        "                        <arg name=\"priority\" type=\"i\" direction=\"in\"/>\n" \
-        "                </method>\n"                                   \
-        "                <method name=\"MakeThreadHighPriorityWithPID\">\n"          \
-        "                        <arg name=\"process\" type=\"t\" direction=\"in\"/>\n" \
-        "                        <arg name=\"thread\" type=\"t\" direction=\"in\"/>\n" \
-        "                        <arg name=\"priority\" type=\"i\" direction=\"in\"/>\n" \
-        "                </method>\n"                                   \
-        "                <method name=\"ResetKnown\"/>\n"               \
-        "                <method name=\"ResetAll\"/>\n"                 \
-        "                <method name=\"Exit\"/>\n"                     \
-        "                <property name=\"RTTimeUSecMax\" type=\"x\" access=\"read\"/>\n" \
-        "                <property name=\"MaxRealtimePriority\" type=\"i\" access=\"read\"/>\n" \
-        "                <property name=\"MinNiceLevel\" type=\"i\" access=\"read\"/>\n" \
-        "        </interface>\n"                                        \
-        "        <interface name=\"org.freedesktop.DBus.Properties\">\n"\
-        "                <method name=\"Get\">"                         \
-        "                       <arg name=\"interface\" direction=\"in\" type=\"s\"/>\n" \
-        "                       <arg name=\"property\" direction=\"in\" type=\"s\"/>\n" \
-        "                       <arg name=\"value\" direction=\"out\" type=\"v\"/>\n" \
-        "                </method>\n"                                   \
-        "        </interface>\n"                                        \
-        "        <interface name=\"org.freedesktop.DBus.Introspectable\">\n" \
-        "                <method name=\"Introspect\">\n"                \
-        "                        <arg name=\"data\" type=\"s\" direction=\"out\"/>\n" \
-        "                </method>\n"                                   \
-        "        </interface>\n"                                        \
-        "</node>\n"
+/* The introspection XML from org.freedesktop.RealtimeKit1.xml (via xml-introspection.S) */
+extern const char introspect_xml[];
 
 /* Similar to assert(), but has side effects, and hence shall never be optimized away, regardless of NDEBUG */
 #define assert_se(expr)                                                 \
@@ -746,26 +706,10 @@ static int thread_reset(pid_t tid) {
         return r;
 }
 
-static char* get_exe_name(pid_t pid, char *exe, size_t len) {
-        char fn[128];
-        ssize_t n;
-
-        assert_se(snprintf(fn, sizeof(fn)-1, "%s/%llu/exe", get_proc_path(), (unsigned long long) pid) < (int) (sizeof(fn)-1));
-        fn[sizeof(fn)-1] = 0;
-
-        if ((n = readlink(fn, exe, len-1)) < 0) {
-                snprintf(exe, len-1, "n/a");
-                exe[len-1] = 0;
-        } else
-                exe[n] = 0;
-
-        return exe;
-}
-
 static int process_set_realtime(struct rtkit_user *u, struct process *p, struct thread *t, unsigned priority) {
         int r;
         struct sched_param param;
-        char user[64], exe[128];
+        char user[64];
 
         if ((int) priority < sched_get_priority_min(sched_policy) ||
             (int) priority > sched_get_priority_max(sched_policy))
@@ -817,10 +761,9 @@ static int process_set_realtime(struct rtkit_user *u, struct process *p, struct
                 goto finish;
         }
 
-        syslog(LOG_INFO, "Successfully made thread %llu of process %llu (%s) owned by '%s' RT at priority %u.\n",
+        syslog(LOG_INFO, "Successfully made thread %llu of process %llu owned by '%s' RT at priority %u.\n",
                (unsigned long long) t->pid,
                (unsigned long long) p->pid,
-               get_exe_name(p->pid, exe, sizeof(exe)),
                get_user_name(u->uid, user, sizeof(user)),
                priority);
 
@@ -835,7 +778,7 @@ finish:
 static int process_set_high_priority(struct rtkit_user *u, struct process *p, struct thread *t, int priority) {
         int r;
         struct sched_param param;
-        char user[64], exe[128];
+        char user[64];
 
         if (priority < PRIO_MIN || priority >= PRIO_MAX)
                 return -EINVAL;
@@ -881,10 +824,9 @@ static int process_set_high_priority(struct rtkit_user *u, struct process *p, st
                 goto finish;
         }
 
-        syslog(LOG_INFO, "Successfully made thread %llu of process %llu (%s) owned by '%s' high priority at nice level %i.\n",
+        syslog(LOG_INFO, "Successfully made thread %llu of process %llu owned by '%s' high priority at nice level %i.\n",
                (unsigned long long) t->pid,
                (unsigned long long) p->pid,
-               get_exe_name(p->pid, exe, sizeof(exe)),
                get_user_name(u->uid, user, sizeof(user)),
                priority);
 
@@ -911,11 +853,9 @@ static void reset_known(void) {
                                     verify_process_starttime(p) >= 0 &&
                                     verify_thread_starttime(p, t) >= 0)
                                         if (thread_reset(t->pid) >= 0) {
-                                                char exe[64];
-                                                syslog(LOG_NOTICE, "Successfully demoted thread %llu of process %llu (%s).\n",
+                                                syslog(LOG_NOTICE, "Successfully demoted thread %llu of process %llu.\n",
                                                        (unsigned long long) t->pid,
-                                                       (unsigned long long) p->pid,
-                                                       get_exe_name(p->pid, exe, sizeof(exe)));
+                                                       (unsigned long long) p->pid);
                                                 n_demoted++;
                                         }
 
@@ -1007,11 +947,9 @@ static int reset_all(void) {
                         if (r == SCHED_FIFO || r == SCHED_RR ||
                             r == (SCHED_FIFO|SCHED_RESET_ON_FORK) || r == (SCHED_RR|SCHED_RESET_ON_FORK))
                                 if (thread_reset((pid_t) tid) >= 0) {
-                                        char exe[64];
-                                        syslog(LOG_NOTICE, "Successfully demoted thread %llu of process %llu (%s).\n",
+                                        syslog(LOG_NOTICE, "Successfully demoted thread %llu of process %llu.\n",
                                                (unsigned long long) tid,
-                                               (unsigned long long) pid,
-                                               get_exe_name(pid, exe, sizeof(exe)));
+                                               (unsigned long long) pid);
                                         n_demoted++;
                                 }
                 }
@@ -1170,12 +1108,14 @@ static int verify_polkit(DBusConnection *c, struct rtkit_user *u, struct process
         DBusMessage *m = NULL, *r = NULL;
         const char *unix_process = "unix-process";
         const char *pid = "pid";
+        const char *uid = "uid";
         const char *start_time = "start-time";
         const char *cancel_id = "";
         uint32_t flags = 0;
         uint32_t pid_u32 = p->pid;
-        uint64_t start_time_u64 = p->starttime;
+        uint32_t uid_u32 = (uint32_t)u->uid;
         DBusMessageIter iter_msg, iter_struct, iter_array, iter_dict, iter_variant;
+        uint64_t start_time_u64 = p->starttime;
         int ret;
         dbus_bool_t authorized = FALSE;
 
@@ -1206,6 +1146,13 @@ static int verify_polkit(DBusConnection *c, struct rtkit_user *u, struct process
         assert_se(dbus_message_iter_close_container(&iter_dict, &iter_variant));
         assert_se(dbus_message_iter_close_container(&iter_array, &iter_dict));
 
+        assert_se(dbus_message_iter_open_container(&iter_array, DBUS_TYPE_DICT_ENTRY, NULL, &iter_dict));
+        assert_se(dbus_message_iter_append_basic(&iter_dict, DBUS_TYPE_STRING, &uid));
+        assert_se(dbus_message_iter_open_container(&iter_dict, DBUS_TYPE_VARIANT, "u", &iter_variant));
+        assert_se(dbus_message_iter_append_basic(&iter_variant, DBUS_TYPE_UINT32, &uid_u32));
+        assert_se(dbus_message_iter_close_container(&iter_dict, &iter_variant));
+        assert_se(dbus_message_iter_close_container(&iter_array, &iter_dict));
+
         assert_se(dbus_message_iter_close_container(&iter_struct, &iter_array));
         assert_se(dbus_message_iter_close_container(&iter_msg, &iter_struct));
 
@@ -1324,7 +1271,7 @@ static DBusHandlerResult dbus_handler(DBusConnection *c, DBusMessage *m, void *u
                 int ret;
 
                 if ((ret = verify_canary_refusal()) < 0) {
-                        assert_se(r = dbus_message_new_error_printf(m, translate_error_forward(ret), strerror(-ret)));
+                        assert_se(r = dbus_message_new_error_printf(m, translate_error_forward(ret), "%s", strerror(-ret)));
                         goto finish;
                 }
 
@@ -1349,17 +1296,17 @@ static DBusHandlerResult dbus_handler(DBusConnection *c, DBusMessage *m, void *u
 
                 if ((ret = lookup_client(&u, &p, &t, c, m, (pid_t)process, (pid_t) thread)) < 0) {
                         syslog(LOG_DEBUG, "Failed to look up client: %s\n", strerror(-ret));
-                        assert_se(r = dbus_message_new_error_printf(m, translate_error_forward(ret), strerror(-ret)));
+                        assert_se(r = dbus_message_new_error_printf(m, translate_error_forward(ret), "%s", strerror(-ret)));
                         goto finish;
                 }
 
                 if ((ret = verify_polkit(c, u, p, "org.freedesktop.RealtimeKit1.acquire-real-time")) < 0) {
-                        assert_se(r = dbus_message_new_error_printf(m, translate_error_forward(ret), strerror(-ret)));
+                        assert_se(r = dbus_message_new_error_printf(m, translate_error_forward(ret), "%s", strerror(-ret)));
                         goto finish;
                 }
 
                 if ((ret = process_set_realtime(u, p, t, priority))) {
-                        assert_se(r = dbus_message_new_error_printf(m, translate_error_forward(ret), strerror(-ret)));
+                        assert_se(r = dbus_message_new_error_printf(m, translate_error_forward(ret), "%s", strerror(-ret)));
                         goto finish;
                 }
 
@@ -1376,7 +1323,7 @@ static DBusHandlerResult dbus_handler(DBusConnection *c, DBusMessage *m, void *u
                 int ret;
 
                 if ((ret = verify_canary_refusal()) < 0) {
-                        assert_se(r = dbus_message_new_error_printf(m, translate_error_forward(ret), strerror(-ret)));
+                        assert_se(r = dbus_message_new_error_printf(m, translate_error_forward(ret), "%s", strerror(-ret)));
                         goto finish;
                 }
 
@@ -1401,17 +1348,17 @@ static DBusHandlerResult dbus_handler(DBusConnection *c, DBusMessage *m, void *u
 
                 if ((ret = lookup_client(&u, &p, &t, c, m, (pid_t)process, (pid_t) thread)) < 0) {
                         syslog(LOG_DEBUG, "Failed to look up client: %s\n", strerror(-ret));
-                        assert_se(r = dbus_message_new_error_printf(m, translate_error_forward(ret), strerror(-ret)));
+                        assert_se(r = dbus_message_new_error_printf(m, translate_error_forward(ret), "%s", strerror(-ret)));
                         goto finish;
                 }
 
                 if ((ret = verify_polkit(c, u, p, "org.freedesktop.RealtimeKit1.acquire-high-priority")) < 0) {
-                        assert_se(r = dbus_message_new_error_printf(m, translate_error_forward(ret), strerror(-ret)));
+                        assert_se(r = dbus_message_new_error_printf(m, translate_error_forward(ret), "%s", strerror(-ret)));
                         goto finish;
                 }
 
                 if ((ret = process_set_high_priority(u, p, t, priority))) {
-                        assert_se(r = dbus_message_new_error_printf(m, translate_error_forward(ret), strerror(-ret)));
+                        assert_se(r = dbus_message_new_error_printf(m, translate_error_forward(ret), "%s", strerror(-ret)));
                         goto finish;
                 }
 
@@ -1454,19 +1401,23 @@ static DBusHandlerResult dbus_handler(DBusConnection *c, DBusMessage *m, void *u
                 if (strcmp(interface, "org.freedesktop.RealtimeKit1") == 0) {
                         assert_se(r = dbus_message_new_method_return(m));
 
-                        if (!handle_dbus_prop_get(property, r) < 0) {
+                        if (handle_dbus_prop_get(property, r) < 0) {
                                 dbus_message_unref(r);
                                 assert_se(r = dbus_message_new_error_printf(
                                           m,
-                                          DBUS_ERROR_UNKNOWN_METHOD,
+                                          DBUS_ERROR_UNKNOWN_PROPERTY,
                                           "Unknown property %s",
                                           property));
                         }
                 } else
-                        return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
+                        assert_se(r = dbus_message_new_error_printf(
+                                m,
+                                DBUS_ERROR_UNKNOWN_PROPERTY,
+                                "Unknown interface %s",
+                                interface));
 
         } else if (dbus_message_is_method_call(m, "org.freedesktop.DBus.Introspectable", "Introspect")) {
-                const char *xml = INTROSPECT_XML;
+                const char *xml = introspect_xml;
 
                 assert_se(r = dbus_message_new_method_return(m));
                 assert_se(dbus_message_append_args(
@@ -1766,8 +1717,7 @@ static int drop_privileges(void) {
         if (do_drop_privileges) {
                 static const cap_value_t cap_values[] = {
                         CAP_SYS_NICE,             /* Needed for obvious reasons */
-                        CAP_DAC_READ_SEARCH,      /* Needed so that we can verify resource limits */
-                        CAP_SYS_PTRACE            /* Needed so that we can read /proc/$$/exe. Linux is weird. */
+                        CAP_DAC_READ_SEARCH      /* Needed so that we can verify resource limits */
                 };
 
                 cap_value_t c, m;
@@ -2266,7 +2216,7 @@ static int parse_command_line(int argc, char *argv[], int *ret) {
                                 break;
 
                         case ARG_INTROSPECT:
-                                fputs(INTROSPECT_XML, stdout);
+                                fputs(introspect_xml, stdout);
                                 *ret = 0;
                                 return 0;
 
